/*
        Algorithm AS 190  Appl. Statist. (1983) Vol.32, No. 2
	Incorporates corrections from Appl. Statist. (1985) Vol.34 (1)

        Evaluates the probability from 0 to q for a studentized
        range having v degrees of freedom and r samples.

        Uses subroutine ALNORM = algorithm AS66.

        Arrays vw and qw store transient values used in the
        quadrature summation.  Node spacing is controlled by
        step.  pcutj and pcutk control truncation.
        Minimum and maximum number of steps are controlled by
        jmin, jmax, kmin and kmax.  Accuracy can be increased
        by use of a finer grid - Increase sizes of arrays vw
        and qw, and jmin, jmax, kmin, kmax and 1/step proportionally.
*/

#include <math.h>
#include <iostream>

double prtrng(double q, double v, double r, short *ifault) {
  long double retprtrng;
  long double vw[31], qw[31];
  long double vmax=120.0;
  long double cv1=0.193064705;
  long double cv2=0.293525326;
  long double cvmax=0.39894228;
  long double cv[5]={0, 0.318309886, -0.268132716e-2, 0.347222222e-2, 0.833333333e-1};
  long double g, gmid, r1, c, h, v2, gstep, pk1, pk2, gk, pk;
  long double w0, pz, x, hj, ehj, pj;
  long double pcutj=0.00003, pcutk=0.0001, step=0.45;
  int  jmin=3, jmax=15, kmin=7, kmax=15;
  int  k, j, jj, jump;

  extern float alnorm(float x, int upper);

  /* check initial values */
  retprtrng=0.0;
  *ifault=0;
  if ((v < 1.0) || (r < 2.0)) *ifault=1;
  if ((q <= 0.0) || (*ifault == 1)) goto ende;

  /* computing constants, local midpoint, adjusting steps */
  g=step * pow(r, -0.2);
  gmid=0.5 * log(r);
  r1=r-1.0;
  c=log(r*g*cvmax);
  if (v > vmax) goto at20;

  h=step * pow(v, -0.5);
  v2=v * 0.5;
  if (v == 1.0) c=cv1;
  if (v == 2.0) c=cv2;
  if (!((v == 1.0) || (v == 2.0)) ) c=sqrt(v2) * cv[1] /(1.0 + ((cv[2]/v2 + cv[3])/ v2 + cv[4]) / v2);
  c=log(c*r*g*h);  

  /* computing integral */
at20:
  gstep=g;
  qw[1]=(-1.0);
  qw[jmax+1]=(-1.0);
  pk1=1.0;
  pk2=1.0;

  for (k=1; k<=kmax; k++) {
    gstep=gstep-g;
  at21:
    gstep=(-1.0)*gstep;
    gk=gmid + gstep;
    pk=0.0;
    if ((pk2 <= pcutk) && (k > kmin)) goto at26;
    w0=c-gk*gk*0.5;
    pz=alnorm(gk, 1);
    x=alnorm(gk-q, 1) - pz;
    if (x > 0.0) pk=exp(w0+r1*log(x));
    if (v > vmax) goto at26;

    jump=(-1)*jmax;
  at22:
    jump=jump + jmax;
    for (j=1; j<=jmax; j++) {
      jj=j+jump;
      if (qw[jj] > 0.0) goto at23;
      hj=h*j;
      if (j < jmax) qw[jj+1]=(-1.0);
      ehj=exp(hj);
      qw[jj]=q*ehj;
      vw[jj]=v*(hj+0.5-ehj*ehj*0.5);
    at23:
      pj=0.0;
      x=alnorm(gk-qw[jj], 1) - pz;
      if (x > 0.0) pj=exp(w0+vw[jj]+r1*log(x));
      pk=pk+pj;
      if (pj > pcutj) goto at24;
      if ((jj > jmin) || (k> kmin)) goto at25;
    at24:
      if (1==2) x=0; // dummy statement;
    }
  at25:
    h=(-1.0)*h;
    if (h < 0.0) goto at22;
at26:
    retprtrng=retprtrng + pk;
    if ((k > kmin) && (pk <= pcutk) && (pk1 <= pcutk)) goto ende;
    pk2=pk1;
    pk1=pk;
    if (gstep > 0.0) goto at21;
  }
  
ende:
  return(retprtrng);;
}

